SystemOrganization addCategory: #ImageSnapshot!

Object subclass: #ImageSnapshot
	instanceVariableNames: 'header memory bigEndian resuming'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ImageSnapshot'!

!ImageSnapshot commentStamp: 'dtl 8/4/2018 12:38' prior: 0!
An ImageSnapshot is a snapshot copy of the state of the running object memory
and its interpreter state. It is equivalent to the data that is written to an image file
when the image is saved, except that it is a Smalltalk object rather than a file
containing image data.

An ImageSnapshot can be written to a file, resulting in an image file like that
produced by a normal image save. It can also be inspected or manipulated like
any other object, so it is potentially useful for other purposes. For example,
an ImageSnapshot might be stored and exchanged in a variety of formats, or
it can be serialized and transmitted directly across a network connection.

The virtual machine is required to implement a primitive to answer a snapshot
of the required object memory and interpreter state at the point of snapshot.
This support is currently available for the classic interpreter VM with V3 images,
and is expected to be extended to Cog and Spur.

Instance variable #bigEndian is the host byte ordering for the snapshot. When
an image is loaded from a traditional image file, the endianness is decoded
by the VM while reading the image header, and is used to determine any
necessary byte and word swapping, as is the case for an image loaded on
a platform with endianness different from that of the saved image. Here we
save the endianness in the ImageSnapshot, and assume that the VM will always
be responsible for byte and word swapping at image load time.

Instance variable #resuming is set based on the return state (on the stack) from
the primitive, and will be false normally but true if the ImageSnapshot instance
has created a saved snapshot instance that is now being resumed.

	| snap |
	snap := ImageSnapshot new.
	snap saveAs: 'newimage.image'.!

ImageSnapshot subclass: #ImageBitmapSnapshot
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ImageSnapshot'!

!ImageBitmapSnapshot commentStamp: 'dtl 8/3/2018 22:40' prior: 0!
An ImageBitmapSnapshot an image snapshot with the object memory represented
as a Bitmap. This is a word-oriented organization suitable for use with an interpreter
simulator.

	| snap |
	snap := ImageBitmapSnapshot new.
	snap saveAs: 'newimage.image'.!

!ImageBitmapSnapshot methodsFor: 'primitive access' stamp: 'dtl 8/4/2018 10:34'!
getSnapshot
	"Answer an array with a snapshot of the object memory as a Bitmap, and with
	an interpreter state array of values suitable for creating an image file header."

	"ImageBitmapSnapshot basicNew getSnapshot"

	<primitive: 'primitiveMemorySnapshotWithHeader'>
	self primitiveFailed! !

!ImageBitmapSnapshot methodsFor: 'saving' stamp: 'dtl 8/4/2018 10:52'!
writeMemoryTo: aStream

	1 to: memory size do: [ :i | | word |
		word := memory at: i.
		bigEndian
			ifFalse: [ "e.g. Intel platforms"
				aStream
					nextPut: (word digitAt:1);
					nextPut: (word digitAt:2);
					nextPut: (word digitAt:3);
					nextPut: (word digitAt:4) ]
			ifTrue: [ "network byte order, early Mac"
				aStream
					nextPut: (word digitAt: 4);
					nextPut: (word digitAt: 3);
					nextPut: (word digitAt: 2);
					nextPut: (word digitAt: 1) ]
	]
! !

!ImageSnapshot class methodsFor: 'saving' stamp: 'dtl 7/30/2018 20:13'!
saveAs: imageFileName

	"ImageSnapshot saveAs: 'newimage.image' "

	^self new saveAs: imageFileName.
! !

!ImageSnapshot methodsFor: 'accessing' stamp: 'dtl 8/4/2018 14:04'!
bigEndian
	^bigEndian! !

!ImageSnapshot methodsFor: 'primitive access' stamp: 'dtl 8/4/2018 10:34'!
getSnapshot
	"Answer an array with a snapshot of the object memory as a ByteArray, and with
	an interpreter state array of values suitable for creating an image file header."

	"ImageSnapshot basicNew getSnapshot"

	<primitive: 'primitiveMemorySnapshotBytesWithHeader'>
	self primitiveFailed! !

!ImageSnapshot methodsFor: 'accessing' stamp: 'dtl 8/4/2018 14:04'!
header
	^header! !

!ImageSnapshot methodsFor: 'initialize-release' stamp: 'dtl 8/4/2018 09:45'!
initialize

	"ImageSnapshot new"

	| snapshot |
	Smalltalk garbageCollect.

	Smalltalk processShutDownList: true.
	snapshot := self getSnapshot.
	Smalltalk processStartUpList: true.

	"The object to be stored in snapshot does not yet exist as of the entry to the
	primitive. If we are resuming from a saved snapshot, it is not possible to access
	that nonexistent object. On resuming, the variable will point to something else.
	A check for the class of that object is sufficient to determine if we are resuming
	from as saved image."

	"In the image snapshot primitive, true will have been pushed onto the stack,
	replacing the normally expected value of an array. In this case, set resuming
	to true so that it is available to startup code."
	(resuming := snapshot == true)
		ifFalse: [ "normal initialization"
			memory := snapshot first.
			header := ImageFileHeader fromValues: snapshot second.
			bigEndian := Smalltalk isBigEndian]
		ifTrue: [ "restarting from saved image"
			self resume ]

! !

!ImageSnapshot methodsFor: 'accessing' stamp: 'dtl 8/4/2018 14:04'!
memory
	^memory! !

!ImageSnapshot methodsFor: 'initialize-release' stamp: 'dtl 7/28/2018 15:42'!
resume
	"Resuming from a saved snapshot. Do startup actions similar to resuming from an
	image saved to disk"

	"Time to reclaim segment files is immediately after a save"
	Smalltalk at: #ImageSegment ifPresent: [:theClass | theClass reclaimObsoleteSegmentFiles].
	Smalltalk setGCParameters.
	Smalltalk clearExternalObjects.
	Smalltalk setPlatformPreferences.
	Project current wakeUpTopWindow.
! !

!ImageSnapshot methodsFor: 'saving' stamp: 'dtl 8/4/2018 10:22'!
saveAs: fileName
	"Save the image to a file in the endianness of this image. In most cases, this
	is little endian byte ordering. Early images saved from Mac will be big endian.
	The VM is responsible for handling endianness at image load time. The VM
	will deduce endianness based on the byte ordering that it detects in the
	saved image header. Note that other possible image loaders might prefer
	to use the bigEndian value of an ImageSnapshot directly."

	"ImageSnapshot new saveAs: 'newimage.image' "
	"ImageBitmapSnapshot new saveAs: 'newimage.image' "

	| fs |
	resuming ifFalse: [  "Do not try to save the same image again when resuming"
		fs := FileStream newFileNamed: fileName.
		fs binary.
		Cursor write showWhile: [
			[ header writeTo: fs littleEndian: bigEndian not.
			self writeMemoryTo: fs]
				ensure: [fs close]]].
! !

!ImageSnapshot methodsFor: 'serializing' stamp: 'dtl 8/4/2018 12:23'!
serializeTo: dataStream
	"Serialize the snapshot to a data stream"

	"Writing to a data stream, saving to file called 'image.dat' "
	" | fs |
	fs := DataStream newFileNamed: 'image.dat'.
	[ ImageSnapshot new serializeTo: fs ]
		ensure: [ fs close]. "

	"Reading the serialized snapshot and storing it as a conventional image file called 'foo.image' "
	" | fs snap |
	[ fs := DataStream oldFileNamed: 'image.dat'.
			snap := fs next.
			snap saveAs: 'foo.image' ]
		ensure: [ fs close .]. "

	resuming ifFalse: [  "Do not try to store again when resuming"
		Cursor write showWhile: [ dataStream nextPut: self]].
! !

!ImageSnapshot methodsFor: 'saving' stamp: 'dtl 8/4/2018 09:58'!
writeMemoryTo: aStream

	aStream nextPutAll: memory
! !
